==> mars-0.9.2.tar.gz: mars-0.9.2/src/mpu/lib/mars_task_event_flag.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <string.h>

#include "mars/mars_task_event_flag.h"
#include "mars/mars_task_types.h"
#include "mars/mars_syscalls.h"
#include "mars/mars_mutex.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

static struct mars_task_event_flag event_flag;

int mars_task_event_flag_initialize(uint64_t event_flag_ea,
				uint8_t direction,
				uint8_t clear_mode)
{
	MARS_CHECK_RET(event_flag_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((event_flag_ea & MARS_TASK_EVENT_FLAG_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);
	MARS_CHECK_RET(direction == MARS_TASK_EVENT_FLAG_HOST_TO_MPU ||
			direction == MARS_TASK_EVENT_FLAG_MPU_TO_HOST ||
			direction == MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET (clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO ||
			clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_MANUAL,
			MARS_ERROR_PARAMS);

	mars_mutex_initialize(event_flag_ea);
	mars_mutex_lock_get(event_flag_ea, (struct mars_mutex *)&event_flag);

	event_flag.mars_context_ea = mars_get_mars_context_ea();
	event_flag.bits = 0;
	event_flag.direction = direction;
	event_flag.clear_mode = clear_mode;
	event_flag.wait_count = 0;

	mars_mutex_unlock_put(event_flag_ea, (struct mars_mutex *)&event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_clear(uint64_t event_flag_ea,
				uint32_t bits)
{
	MARS_CHECK_RET(event_flag_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((event_flag_ea & MARS_TASK_EVENT_FLAG_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	mars_mutex_lock_get(event_flag_ea, (struct mars_mutex *)&event_flag);

	/* clear the necessary bits */
	event_flag.bits &= ~bits;

	mars_mutex_unlock_put(event_flag_ea, (struct mars_mutex *)&event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_set(uint64_t event_flag_ea,
				uint32_t bits)
{
	MARS_CHECK_RET(event_flag_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((event_flag_ea & MARS_TASK_EVENT_FLAG_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	int i;

	mars_mutex_lock_get(event_flag_ea, (struct mars_mutex *)&event_flag);

	/* check for valid direction */
	MARS_CHECK_CLEANUP_RET(
		event_flag.direction == MARS_TASK_EVENT_FLAG_MPU_TO_HOST ||
		event_flag.direction == MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
		mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag),
		MARS_ERROR_STATE);

	/* set the necessary bits */
	event_flag.bits |= bits;

	/* save current set bits */
	bits = event_flag.bits;

	/* search through wait list for tasks to be signalled */
	for (i = 0; i < event_flag.wait_count; i++) {
		/* check condition based on wait mode */
		switch (event_flag.wait_mask_mode[i]) {
		case MARS_TASK_EVENT_FLAG_MASK_OR:
			if ((bits & event_flag.wait_mask[i]) == 0)
				continue;
			break;
		case MARS_TASK_EVENT_FLAG_MASK_AND:
			if ((bits & event_flag.wait_mask[i])
				!= event_flag.wait_mask[i])
				continue;
			break;
		}

		/* clear event if clear mode is auto */
		if (event_flag.clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO)
			event_flag.bits &= ~event_flag.wait_mask[i];

		/* signal the waiting tasks */
		mars_signal_send(event_flag.wait_id[i]);

		/* flush id from wait list */
		event_flag.wait_count--;
		memmove(&event_flag.wait_id[i],
			&event_flag.wait_id[i + 1],
			sizeof(uint16_t) * (event_flag.wait_count - i));
		memmove(&event_flag.wait_mask[i],
			&event_flag.wait_mask[i + 1],
			sizeof(uint32_t) * (event_flag.wait_count - i));
		memmove(&event_flag.wait_mask_mode[i],
			&event_flag.wait_mask_mode[i + 1],
			sizeof(uint8_t) * (event_flag.wait_count - i));
		i--;
	}

	mars_mutex_unlock_put(event_flag_ea, (struct mars_mutex *)&event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_wait(uint64_t event_flag_ea,
				uint32_t mask,
				uint8_t mask_mode)
{
	MARS_CHECK_RET(event_flag_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET(mask_mode == MARS_TASK_EVENT_FLAG_MASK_OR ||
			mask_mode == MARS_TASK_EVENT_FLAG_MASK_AND,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET((event_flag_ea & MARS_TASK_EVENT_FLAG_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	int wait = 0;
	struct mars_task_context *task;

	task = (struct mars_task_context *)mars_get_workload_context();

	/* make sure task context has a context save area */
	MARS_CHECK_RET(task->context_save_size && task->context_save_area,
			MARS_ERROR_FORMAT);

	mars_mutex_lock_get(event_flag_ea, (struct mars_mutex *)&event_flag);

	/* check if event flag wait limit reached */
	MARS_CHECK_CLEANUP_RET(
		event_flag.wait_count < MARS_TASK_EVENT_FLAG_WAIT_MAX,
		mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag),
		MARS_ERROR_LIMIT);

	/* check for valid direction */
	MARS_CHECK_CLEANUP_RET(
		event_flag.direction == MARS_TASK_EVENT_FLAG_HOST_TO_MPU ||
		event_flag.direction == MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
		mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag),
		MARS_ERROR_STATE);

	/* check condition based on wait mode */
	switch (mask_mode) {
	case MARS_TASK_EVENT_FLAG_MASK_OR:
		if ((event_flag.bits & mask) == 0)
			wait = 1;
		break;
	case MARS_TASK_EVENT_FLAG_MASK_AND:
		if ((event_flag.bits & mask) != mask)
			wait = 1;
		break;
	}

	if (wait) {
		/* add id to wait list */
		event_flag.wait_id[event_flag.wait_count]
				= mars_get_workload_id();
		event_flag.wait_mask[event_flag.wait_count]
				= mask;
		event_flag.wait_mask_mode[event_flag.wait_count]
				= mask_mode;
		event_flag.wait_count++;

		mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag);

		/* wait for signal */
		mars_signal_wait();

		return MARS_SUCCESS;
	}

	/* clear event if clear mode is auto */
	if (event_flag.clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO)
		event_flag.bits &= ~mask;

	mars_mutex_unlock_put(event_flag_ea, (struct mars_mutex *)&event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_try_wait(uint64_t event_flag_ea,
				uint32_t mask,
				uint8_t mask_mode)
{
	MARS_CHECK_RET(event_flag_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET(mask_mode == MARS_TASK_EVENT_FLAG_MASK_OR ||
			mask_mode == MARS_TASK_EVENT_FLAG_MASK_AND,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET((event_flag_ea & MARS_TASK_EVENT_FLAG_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	mars_mutex_lock_get(event_flag_ea, (struct mars_mutex *)&event_flag);

	/* check for valid direction */
	MARS_CHECK_CLEANUP_RET(
		event_flag.direction == MARS_TASK_EVENT_FLAG_HOST_TO_MPU ||
		event_flag.direction == MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
		mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag),
		MARS_ERROR_STATE);

	/* check condition based on wait mode */
	switch (mask_mode) {
	case MARS_TASK_EVENT_FLAG_MASK_OR:
		if ((event_flag.bits & mask) == 0) {
			mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag);
			return MARS_ERROR_BUSY;
		}
		break;
	case MARS_TASK_EVENT_FLAG_MASK_AND:
		if ((event_flag.bits & mask) != mask) {
			mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag);
			return MARS_ERROR_BUSY;
		}
		break;
	}

	/* clear event if clear mode is auto */
	if (event_flag.clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO)
		event_flag.bits &= ~mask;

	mars_mutex_unlock_put(event_flag_ea, (struct mars_mutex *)&event_flag);

	return MARS_SUCCESS;
}

==> mars-08.06.26.tar.gz: mars-08.06.26/src/mpu/lib/mars_task_event_flag.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <string.h>

#include "mars/mars_task_event_flag.h"
#include "mars/mars_workload_types.h"
#include "mars/mars_syscalls.h"
#include "mars/mars_mutex.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

static struct mars_task_event_flag event_flag;

int mars_task_event_flag_initialize(uint64_t event_flag_ea,
				uint8_t direction,
				uint8_t clear_mode)
{
	MARS_CHECK_RET(event_flag_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((event_flag_ea & MARS_TASK_EVENT_FLAG_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);
	MARS_CHECK_RET(direction == MARS_TASK_EVENT_FLAG_HOST_TO_MPU ||
			direction == MARS_TASK_EVENT_FLAG_MPU_TO_HOST ||
			direction == MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET (clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO ||
			clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_MANUAL,
			MARS_ERROR_PARAMS);

	mars_mutex_initialize(event_flag_ea);
	mars_mutex_lock_get(event_flag_ea, (struct mars_mutex *)&event_flag);

	event_flag.mars_context_ea = mars_get_mars_context_ea();
	event_flag.bits = 0;
	event_flag.direction = direction;
	event_flag.clear_mode = clear_mode;
	event_flag.wait_count = 0;

	mars_mutex_unlock_put(event_flag_ea, (struct mars_mutex *)&event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_clear(uint64_t event_flag_ea,
				uint32_t bits)
{
	MARS_CHECK_RET(event_flag_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((event_flag_ea & MARS_TASK_EVENT_FLAG_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	mars_mutex_lock_get(event_flag_ea, (struct mars_mutex *)&event_flag);

	/* clear the necessary bits */
	event_flag.bits &= ~bits;

	mars_mutex_unlock_put(event_flag_ea, (struct mars_mutex *)&event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_set(uint64_t event_flag_ea,
				uint32_t bits)
{
	MARS_CHECK_RET(event_flag_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((event_flag_ea & MARS_TASK_EVENT_FLAG_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	int i;

	mars_mutex_lock_get(event_flag_ea, (struct mars_mutex *)&event_flag);

	/* check for valid direction */
	MARS_CHECK_CLEANUP_RET(
		event_flag.direction == MARS_TASK_EVENT_FLAG_MPU_TO_HOST ||
		event_flag.direction == MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
		mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag),
		MARS_ERROR_STATE);

	/* set the necessary bits */
	event_flag.bits |= bits;

	/* save current set bits */
	bits = event_flag.bits;

	/* search through wait list for tasks to be signalled */
	for (i = 0; i < event_flag.wait_count; i++) {
		/* check condition based on wait mode */
		switch (event_flag.wait_mask_mode[i]) {
		case MARS_TASK_EVENT_FLAG_MASK_OR:
			if ((bits | event_flag.wait_mask[i]) == 0)
				continue;
			break;
		case MARS_TASK_EVENT_FLAG_MASK_AND:
			if ((bits & event_flag.wait_mask[i])
				!= event_flag.wait_mask[i])
				continue;
			break;
		}

		/* clear event if clear mode is auto */
		if (event_flag.clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO)
			event_flag.bits &= ~event_flag.wait_mask[i];

		/* signal the waiting tasks */
		mars_signal_send(event_flag.wait_id[i]);

		/* flush id from wait list */
		event_flag.wait_count--;
		memmove(&event_flag.wait_id[i],
			&event_flag.wait_id[i + 1],
			sizeof(uint16_t) * (event_flag.wait_count - i));
		memmove(&event_flag.wait_mask[i],
			&event_flag.wait_mask[i + 1],
			sizeof(uint32_t) * (event_flag.wait_count - i));
		memmove(&event_flag.wait_mask_mode[i],
			&event_flag.wait_mask_mode[i + 1],
			sizeof(uint8_t) * (event_flag.wait_count - i));
		i--;
	}

	mars_mutex_unlock_put(event_flag_ea, (struct mars_mutex *)&event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_wait(uint64_t event_flag_ea,
				uint32_t mask,
				uint8_t mask_mode)
{
	MARS_CHECK_RET(event_flag_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET(mask_mode == MARS_TASK_EVENT_FLAG_MASK_OR ||
			mask_mode == MARS_TASK_EVENT_FLAG_MASK_AND,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET((event_flag_ea & MARS_TASK_EVENT_FLAG_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	int wait = 0;
	struct mars_task_context *task;

	/* make sure task context has a context save area */
	task = &mars_get_workload_context()->type_context.task;
	MARS_CHECK_RET(task->context_save_size && task->context_save_area,
			MARS_ERROR_FORMAT);

	mars_mutex_lock_get(event_flag_ea, (struct mars_mutex *)&event_flag);

	/* check if event flag wait limit reached */
	MARS_CHECK_CLEANUP_RET(
		event_flag.wait_count < MARS_TASK_EVENT_FLAG_WAIT_MAX,
		mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag),
		MARS_ERROR_LIMIT);

	/* check for valid direction */
	MARS_CHECK_CLEANUP_RET(
		event_flag.direction == MARS_TASK_EVENT_FLAG_HOST_TO_MPU ||
		event_flag.direction == MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
		mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag),
		MARS_ERROR_STATE);

	/* check condition based on wait mode */
	switch (mask_mode) {
	case MARS_TASK_EVENT_FLAG_MASK_OR:
		if ((event_flag.bits | mask) == 0)
			wait = 1;
		break;
	case MARS_TASK_EVENT_FLAG_MASK_AND:
		if ((event_flag.bits & mask) != mask)
			wait = 1;
		break;
	}

	if (wait) {
		/* add id to wait list */
		event_flag.wait_id[event_flag.wait_count]
				= mars_get_workload_id();
		event_flag.wait_mask[event_flag.wait_count]
				= mask;
		event_flag.wait_mask_mode[event_flag.wait_count]
				= mask_mode;
		event_flag.wait_count++;

		mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag);

		/* wait for signal */
		mars_signal_wait();

		return MARS_SUCCESS;
	}

	mars_mutex_unlock_put(event_flag_ea, (struct mars_mutex *)&event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_try_wait(uint64_t event_flag_ea,
				uint32_t mask,
				uint8_t mask_mode)
{
	MARS_CHECK_RET(event_flag_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET(mask_mode == MARS_TASK_EVENT_FLAG_MASK_OR ||
			mask_mode == MARS_TASK_EVENT_FLAG_MASK_AND,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET((event_flag_ea & MARS_TASK_EVENT_FLAG_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	mars_mutex_lock_get(event_flag_ea, (struct mars_mutex *)&event_flag);

	/* check for valid direction */
	MARS_CHECK_CLEANUP_RET(
		event_flag.direction == MARS_TASK_EVENT_FLAG_HOST_TO_MPU ||
		event_flag.direction == MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
		mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag),
		MARS_ERROR_STATE);

	/* check condition based on wait mode */
	switch (mask_mode) {
	case MARS_TASK_EVENT_FLAG_MASK_OR:
		if ((event_flag.bits | mask) == 0) {
			mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag);
			return MARS_ERROR_BUSY;
		}
		break;
	case MARS_TASK_EVENT_FLAG_MASK_AND:
		if ((event_flag.bits & mask) != mask) {
			mars_mutex_unlock_put(event_flag_ea,
				(struct mars_mutex *)&event_flag);
			return MARS_ERROR_BUSY;
		}
		break;
	}

	/* clear event if clear mode is auto */
	if (event_flag.clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO)
		event_flag.bits &= ~mask;

	mars_mutex_unlock_put(event_flag_ea, (struct mars_mutex *)&event_flag);

	return MARS_SUCCESS;
}
